from openfisca_us.model_api import *


class state_income_tax_payroll_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax payroll tax deduction"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        dependent = person("is_tax_unit_dependent", period)
        TAXES = [
            "employee_social_security_tax",
            "employee_medicare_tax",
            "self_employment_tax",
        ]
        taxes = add(person, period, TAXES)
        state = person.household("state_code_str", period)
        cap = parameters(period).states.tax.income.deductions.payroll_tax[
            state
        ]
        capped_payroll_tax = min_(taxes, cap)
        eligible_capped_payroll_tax = ~dependent * capped_payroll_tax
        return tax_unit.sum(eligible_capped_payroll_tax)



from openfisca_us.model_api import *


class c17000(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Medical expense deduction"
    reference = "https://www.law.cornell.edu/uscode/text/26/213"
    unit = USD
    documentation = "Sch A: Medical expenses deducted (component of pre-limitation c21060 total)"

    def formula(tax_unit, period, parameters):
        medical = parameters(period).irs.deductions.itemized.medical
        has_aged = (tax_unit("age_head", period) >= 65) | (
            tax_unit("tax_unit_is_joint", period)
            & (tax_unit("age_spouse", period) >= 65)
        )
        medical_floor_ratio = (
            medical.floor.base + has_aged * medical.floor.aged_addition
        )
        medical_floor = medical_floor_ratio * max_(
            tax_unit("adjusted_gross_income", period), 0
        )
        return max_(
            0,
            tax_unit("filer_e17500", period) - medical_floor,
        )


class c18300(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "SALT deduction"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/164"
    documentation = "Sch A: State and local taxes plus real estate taxes deducted (component of pre-limitation c21060 total)"

    def formula(tax_unit, period, parameters):
        c18400 = max_(tax_unit("filer_e18400", period), 0)
        c18500 = tax_unit("filer_e18500", period)
        salt = parameters(period).irs.deductions.itemized.salt_and_real_estate
        cap = salt.cap[tax_unit("filing_status", period)]
        return min_(c18400 + c18500, cap)


class c19200(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Interest deduction"
    reference = "https://www.law.cornell.edu/uscode/text/26/163"
    unit = USD
    documentation = (
        "Sch A: Interest deducted (component of pre-limitation c21060 total)"
    )

    def formula(tax_unit, period, parameters):
        return tax_unit("filer_e19200", period)


class c19700(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Charitable deduction"
    reference = "https://www.law.cornell.edu/uscode/text/26/170"
    unit = USD
    documentation = "Sch A: Charity contributions deducted (component of pre-limitation c21060 total)"

    def formula(tax_unit, period, parameters):
        charity = parameters(period).irs.deductions.itemized.charity
        posagi = tax_unit("posagi", period)
        lim30 = min_(
            charity.ceiling.non_cash * posagi,
            tax_unit("filer_e20100", period),
        )
        c19700 = min_(
            charity.ceiling.all * posagi,
            lim30 + tax_unit("filer_e19800", period),
        )
        return max_(c19700, 0)


class c20500(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Casualty deduction"
    reference = "https://www.law.cornell.edu/uscode/text/26/165"
    unit = USD
    documentation = "Sch A: Net casualty or theft loss deducted (component of pre-limitation c21060 total)"

    def formula(tax_unit, period, parameters):
        casualty = parameters(period).irs.deductions.itemized.casualty
        floor = casualty.floor * tax_unit("posagi", period)
        deduction = max_(0, tax_unit("filer_g20500", period) - floor)
        return deduction * (1 - casualty.haircut)


class c20800(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Miscellaneous deductions"
    unit = USD
    documentation = "Sch A: Net limited miscellaneous deductions deducted (component of pre-limitation c21060 total)"

    def formula(tax_unit, period, parameters):
        misc = parameters(period).irs.deductions.itemized.misc
        floor = misc.floor * tax_unit("posagi", period)
        deduction = max_(0, tax_unit("filer_e20400", period) - floor)
        return deduction * (1 - misc.haircut)



from openfisca_us.model_api import *


class c33200(Variable):
    value_type = float
    entity = TaxUnit
    label = "Credit for child and dependent care expenses"
    unit = USD
    documentation = "From form 2441, before refundability checks"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21"

    def formula(tax_unit, period, parameters):
        cdcc = parameters(period).irs.credits.cdcc
        eligible_deps = min_(tax_unit("f2441", period), cdcc.eligibility.max)
        max_credit = eligible_deps * cdcc.max
        c32800 = max_(0, min_(tax_unit("filer_e32800", period), max_credit))
        filing_status = tax_unit("filing_status", period)
        person = tax_unit.members
        is_head = person("is_tax_unit_head", period)
        earnings = person("earned", period)
        is_spouse = person("is_tax_unit_spouse", period)
        head_earnings = tax_unit.sum(is_head * earnings)
        spouse_earnings = tax_unit.sum(is_spouse * earnings)
        lowest_earnings = where(
            filing_status == filing_status.possible_values.JOINT,
            min_(head_earnings, spouse_earnings),
            head_earnings,
        )
        c33000 = max_(0, min_(c32800, lowest_earnings))
        adjusted_gross_income = tax_unit("adjusted_gross_income", period)
        tratio = 0.01 * max_(
            (
                (adjusted_gross_income - cdcc.phaseout.start)
                * cdcc.phaseout.rate
            ),
            0,
        )
        crate = max_(
            cdcc.phaseout.min,
            cdcc.phaseout.max
            - min_(
                cdcc.phaseout.max - cdcc.phaseout.min,
                tratio,
            ),
        )
        tratio2 = max_(
            (
                (adjusted_gross_income - cdcc.phaseout.second_start)
                * cdcc.phaseout.rate
                / 1e2
            ),
            0,
        )
        crate_if_over_second_threshold = max_(
            0, cdcc.phaseout.min - min_(cdcc.phaseout.min, tratio2)
        )
        crate = where(
            adjusted_gross_income > cdcc.phaseout.second_start,
            crate_if_over_second_threshold,
            crate,
        )

        return c33000 * crate



from openfisca_us.model_api import *


class childcare_expenses(Variable):
    value_type = float
    entity = SPMUnit
    label = "Child care expenses"
    documentation = "Out of pocket spending on childcare"
    definition_period = YEAR
    unit = USD



from openfisca_us.model_api import *


class state_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax before refundable credits"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        tax_before_credits = tax_unit(
            "state_income_tax_before_credits", period
        )
        non_refundable_credits = tax_unit(
            "state_income_tax_non_refundable_credits", period
        )
        return tax_before_credits - non_refundable_credits



from openfisca_us.model_api import *


class state_income_tax_exempt_limit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Income limit to be exempt from state income tax"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).states.tax.income.exempt
        state = tax_unit.household("state_code_str", period)
        filing_status = tax_unit("filing_status", period)
        dependents = tax_unit("tax_unit_dependents", period)
        base_limit = p.limit[state][filing_status]
        dep_limit = dependents * p.dependent[state][filing_status]
        return base_limit + dep_limit



from openfisca_us.model_api import *


class state_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "State Earned Income Tax Credit"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        match = parameters(
            period
        ).states.tax.income.credits.refundable.eitc.percent_match
        state = tax_unit.household("state_code_str", period)
        return tax_unit("eitc", period) * match[state]



from openfisca_us.model_api import *


class state_income_tax_personal_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax personal exemption"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        exemptions = parameters(period).states.tax.income.exemptions
        state = tax_unit.household("state_code_str", period)
        filing_status = tax_unit("filing_status", period)
        return exemptions.personal[state][filing_status]



from openfisca_us.model_api import *


class is_self_employed(Variable):
    value_type = bool
    entity = Person
    label = "Is self-employed"
    documentation = "Whether this person is self-employed."
    definition_period = YEAR



from openfisca_us.model_api import *


class medicaid_income_threshold(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Medicaid FPL threshold"
    documentation = "Maximum income as a percentage of the federal poverty line to qualify for Medicaid"
    unit = "percent"

    def formula(person, period, parameters):
        state_code = person.household("state_code_str", period)
        person_type = person("medicaid_person_type", period)
        income_threshold = parameters(period).hhs.medicaid.income.limit
        return income_threshold[state_code][person_type]



from openfisca_us.model_api import *


class is_wic_at_nutritional_risk(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Assessed as being at nutritional risk for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "At nutritional risk for WIC"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#b_8"



from openfisca_us.model_api import *


class household_vehicles_owned(Variable):
    value_type = float
    entity = Household
    label = "Vehicles owned"
    unit = USD
    documentation = "Number of vehicles owned by the household"
    definition_period = YEAR



from openfisca_us.model_api import *


class income_tax_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Refundable tax credits"
    documentation = "Total refundable income tax credits"
    unit = USD

    def formula(tax_unit, period, parameters):
        credits = parameters(period).irs.credits.refundable
        return add(tax_unit, period, credits)



from openfisca_us.model_api import *


class state_income_tax_blind_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax blind exemption"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        amount = parameters(period).states.tax.income.exemptions.blind
        state = tax_unit.household("state_code_str", period)
        head_eligible = tax_unit("blind_head", period)
        spouse_eligible = tax_unit("blind_spouse", period)
        return amount[state] * (
            head_eligible.astype(int) + spouse_eligible.astype(int)
        )



from openfisca_us.model_api import *


class state_income_tax_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "State non-refundable income tax credits"
    unit = USD
    definition_period = YEAR

    formula = sum_of_variables(["state_limited_income_tax_credit"])



from openfisca_us.model_api import *


class state_income_tax_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax refundable credits"
    unit = USD
    definition_period = YEAR

    formula = sum_of_variables(["state_eitc", "state_dependent_credit"])



from openfisca_us.model_api import *


class state_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax before credits"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        pit = parameters(period).states.tax.income
        state = tax_unit.household("state_code_str", period)
        rate = pit.rates[state]
        exempt = tax_unit("is_state_income_tax_exempt", period)
        return tax_unit("state_taxable_income", period) * rate * ~exempt



from openfisca_us.model_api import *


class ssi_countable_income(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Supplemental Security Income countable income"
    label = "SSI countable income"
    unit = USD

    def formula(person, period, parameters):
        earned = person("ssi_earned_income", period) / 12
        unearned = person("ssi_unearned_income", period) / 12
        deemed = person("ssi_deemed_income", period) / 12
        # Deduct exclusions.
        exclusions = parameters(period).ssa.ssi.income.exclusions
        # Subtract from unearned income first.
        unearned_exclusion = min_(exclusions.general, unearned)
        countable_unearned = unearned - unearned_exclusion
        # Remaining general exclusion is treated as an earned income exclusion.
        remaining_general_exclusion = exclusions.general - unearned_exclusion
        earned_exclusion = exclusions.earned + remaining_general_exclusion
        # Subtract the percentage of earned income above the flat exclusion.
        earned_after_flat_exclusion = max_(earned - earned_exclusion, 0)
        countable_earned_share = 1 - exclusions.earned_share
        countable_earned = earned_after_flat_exclusion * countable_earned_share
        return (
            countable_unearned + countable_earned + deemed
        ) * MONTHS_IN_YEAR



from openfisca_us.model_api import *


class c21060(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Gross itemized deductions"
    unit = USD
    documentation = (
        "Itemized deductions before phase-out (zero for non-itemizers)"
    )

    formula = sum_of_variables(
        ["c17000", "c18300", "c19200", "c19700", "c20500", "c20800"]
    )



from openfisca_us.model_api import *


class state_income_tax_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax deductions"
    unit = USD
    definition_period = YEAR

    formula = sum_of_variables(
        [
            "state_income_tax_payroll_tax_deduction",
            "state_income_tax_rental_deduction",
        ]
    )



from openfisca_us.model_api import *


class ssi_deemed_income(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Supplemental Security Income deemed income"
    label = "SSI deemed income"
    unit = USD
    reference = "https://ncler.acl.gov/NCLER/media/NCLER/documents/SSI-Deeming-Chapter-Summary.pdf"



from openfisca_us.model_api import *


class ssi(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Supplemental Security Income"
    label = "Supplemental Security Income"
    unit = USD

    def formula(person, period, parameters):
        abd = person("is_ssi_aged_blind_disabled", period)
        countable_resources = person("ssi_countable_resources", period)
        ssi = parameters(period).ssa.ssi
        # Only individual is modeled currently.
        resource_limit = ssi.eligibility.resources.limit.individual
        meets_resource_test = countable_resources <= resource_limit
        # Calculate amount.
        amount = ssi.amount.individual * MONTHS_IN_YEAR
        countable_income = person("ssi_countable_income", period)
        amount_if_eligible = max_(amount - countable_income, 0)
        return abd * meets_resource_test * amount_if_eligible



from openfisca_us.model_api import *


class MedicaidPersonType(Enum):
    ADULT_WITHOUT_DEPENDENT = "Adult without dependent"
    ADULT_WITH_DEPENDENT = "Adult with dependent"
    PREGNANT = "Pregnant adult"
    CHILD_AGE_0 = "Child age 0"
    CHILD_AGE_1_5 = "Child age 1 to 5"
    CHILD_AGE_6_18 = "Child age 6 to 18"
    CHILD_AGE_19_20 = "Adult under 21 (covered as child)"


class medicaid_person_type(Variable):
    value_type = Enum
    possible_values = MedicaidPersonType
    default_value = MedicaidPersonType.ADULT_WITHOUT_DEPENDENT
    entity = Person
    definition_period = YEAR
    label = "Medicaid person type"
    documentation = "Person type for Medicaid"

    def formula(person, period, parameters):
        ma = parameters(period).hhs.medicaid
        # Get the person's age.
        age = person("age", period)
        # Get the existence of dependents, as defined by people 18 or younger.
        has_dependents = person.spm_unit.any(
            person.spm_unit.members("age", period) < ma.dependent_age
        )
        under_21_qualifies = ma.under_21_qualifies_as_child
        state = person.household("state_code_str", period)
        state_has_under_21_child_category = under_21_qualifies[state] > 0
        is_pregnant = person("is_pregnant", period)
        days_postpartum = person("count_days_postpartum", period)
        max_postpartum_days = ma.postpartum_coverage[state]
        is_covered_as_pregnant = is_pregnant | (
            days_postpartum < max_postpartum_days
        )
        return select(
            [
                is_covered_as_pregnant,
                age == 0,
                age < 6,
                age < 19,
                (age < 21) & state_has_under_21_child_category,
                has_dependents,
                True,
            ],
            [
                MedicaidPersonType.PREGNANT,
                MedicaidPersonType.CHILD_AGE_0,
                MedicaidPersonType.CHILD_AGE_1_5,
                MedicaidPersonType.CHILD_AGE_6_18,
                MedicaidPersonType.CHILD_AGE_19_20,
                MedicaidPersonType.ADULT_WITH_DEPENDENT,
                MedicaidPersonType.ADULT_WITHOUT_DEPENDENT,
            ],
        )



from openfisca_us.model_api import *


class qbided(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "QBI deduction"
    documentation = "Qualified Business Income (QBI) deduction"
    reference = "https://www.law.cornell.edu/uscode/text/26/199A"
    unit = USD

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        QBINC_ELEMENTS = [
            "filer_e00900",
            "filer_e26270",
            "filer_e02100",
            "filer_e27200",
        ]
        qbinc = max_(0, add(tax_unit, period, QBINC_ELEMENTS))
        qbid = parameters(period).irs.deductions.qualified_business_interest
        lower_threshold = qbid.threshold.lower[filing_status]
        upper_threshold = lower_threshold + qbid.threshold.gap[filing_status]
        pre_qbid_taxinc = tax_unit("pre_qbid_taxinc", period)
        under_lower_threshold = pre_qbid_taxinc < lower_threshold
        between_thresholds = ~under_lower_threshold & (
            pre_qbid_taxinc < upper_threshold
        )
        above_upper_threshold = ~under_lower_threshold & ~between_thresholds
        income_is_qualified = tax_unit("pt_sstb_income", period)

        # Wage/capital limitations
        w2_wages = tax_unit("pt_binc_w2_wages", period)
        business_property = tax_unit("pt_ubia_property", period)
        wage_cap = w2_wages * qbid.cap.w2_wages.rate
        alt_cap = (
            w2_wages * qbid.cap.w2_wages.alt_rate
            + business_property * qbid.cap.business_property.rate
        )
        fraction_of_gap_passed = (
            pre_qbid_taxinc - lower_threshold
        ) / qbid.threshold.gap[filing_status]
        fraction_of_gap_unused = (
            upper_threshold - pre_qbid_taxinc
        ) / qbid.threshold.gap[filing_status]

        # Adjustments for qualified income under the upper threshold
        qbi_between_threshold_multiplier = where(
            income_is_qualified & between_thresholds,
            fraction_of_gap_unused,
            1.0,
        )
        max_qbid = (
            qbinc * qbid.pass_through_rate * qbi_between_threshold_multiplier
        )
        full_cap = max_(wage_cap, alt_cap) * qbi_between_threshold_multiplier

        # Adjustment for QBID where income is between the main thresholds
        adjustment = fraction_of_gap_passed * (max_qbid - full_cap)

        qbid_amount = select(
            (
                under_lower_threshold,
                between_thresholds,
                above_upper_threshold,
            ),
            (
                max_qbid,
                max_qbid - adjustment,
                where(income_is_qualified, 0, min_(max_qbid, full_cap)),
            ),
        )

        # Apply taxable income cap
        net_cg = add(tax_unit, period, ["filer_e00650", "c01000"])
        taxinc_cap = qbid.pass_through_rate * max_(0, pre_qbid_taxinc - net_cg)
        return min_(qbid_amount, taxinc_cap)



from openfisca_us.model_api import *


class ssi_unearned_income(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    documentation = "Supplemental Security Income unearned income"
    label = "SSI unearned income"
    unit = USD

    def formula(person, period, parameters):
        sources = parameters(period).ssa.ssi.income.sources.unearned
        return add(person, period, sources)



from openfisca_us.model_api import *


class ctc_refundable_individual_maximum(Variable):
    value_type = float
    entity = Person
    label = "Maximum refundable CTC"
    unit = USD
    documentation = "The maximum refundable CTC for this person."
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/24#a",
        "https://www.law.cornell.edu/uscode/text/26/24#h",
        "https://www.law.cornell.edu/uscode/text/26/24#i",
    )

    def formula(person, period, parameters):
        return person("ctc_individual_maximum", period)

    def formula_2018(person, period, parameters):
        amount = person("ctc_individual_maximum", period)
        ctc = parameters(period).irs.credits.ctc
        return min_(amount, ctc.refundable.individual_max)

    def formula_2021(person, period, parameters):
        return person("ctc_child_individual_maximum", period)

    formula_2022 = formula_2018

    formula_2026 = formula


class ctc_limiting_tax_liability(Variable):
    value_type = float
    entity = TaxUnit
    label = "CTC-limiting tax liability"
    unit = USD
    documentation = "The tax liability used to determine the maximum amount of the non-refundable CTC."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        non_refundable_credits = parameters(period).irs.credits.non_refundable
        total_credits = sum(
            [
                tax_unit(credit, period)
                for credit in non_refundable_credits
                if credit != "non_refundable_ctc"
            ]
        )
        return tax_unit("income_tax_before_credits", period) - total_credits


class refundable_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Refundable CTC"
    unit = USD
    documentation = "The portion of the Child Tax Credit that is refundable."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#d"

    def formula(tax_unit, period, parameters):

        # This line corresponds to "the credit which would be allowed under this section [the CTC section]"
        # without regard to this subsection [the refundability section] and the limitation under
        # section 26(a) [the section that limits the amount of the non-refundable CTC to tax liability].
        # This is the full CTC. This is then limited to the maximum refundable amount per child as per the
        # TCJA provision.

        ctc = parameters(period).irs.credits.ctc

        total_ctc = tax_unit("ctc", period)
        maximum_refundable_ctc = min_(
            aggr(tax_unit, period, ["ctc_refundable_individual_maximum"]),
            total_ctc,
        )

        # The other part of the "lesser of" statement is: "the amount by which [the non-refundable CTC]
        # would increase if [tax liability] increased by tax_increase", where tax_increase is the greater of:
        # - the phase-in amount
        # - Social Security tax minus the EITC
        # First, we find tax_increase:

        earnings = aggr(tax_unit, period, ["earned_income"])
        earnings_over_threshold = max_(
            0, earnings - ctc.refundable.phase_in.threshold
        )
        relevant_earnings = (
            earnings_over_threshold * ctc.refundable.phase_in.rate
        )

        # Compute "Social Security taxes" as defined in the US Code for the ACTC.
        # This includes OASDI and Medicare payroll taxes, as well as half
        # of self-employment taxes.
        PERSON_VARIABLES = [
            "employee_social_security_tax",
            "employee_medicare_tax",
            "e09800",  # Unreported payroll tax.
        ]
        PERSON_VARIABLES_SUBTRACT = ["e11200"]  # Excess payroll tax withheld.
        TAX_UNIT_VARIABLES = [
            "c03260",  # Deductible portion of the self-employed tax.
            "additional_medicare_tax",
        ]
        social_security_tax = (
            aggr(tax_unit, period, PERSON_VARIABLES)
            + add(tax_unit, period, TAX_UNIT_VARIABLES)
            - aggr(tax_unit, period, PERSON_VARIABLES_SUBTRACT)
        )
        eitc = tax_unit("eitc", period)
        social_security_excess = max_(0, social_security_tax - eitc)

        tax_increase = max_(relevant_earnings, social_security_excess)

        limiting_tax = tax_unit("ctc_limiting_tax_liability", period)
        ctc_capped_by_tax = min_(total_ctc, limiting_tax)
        ctc_capped_by_increased_tax = min_(
            total_ctc, limiting_tax + tax_increase
        )
        amount_ctc_would_increase = (
            ctc_capped_by_increased_tax - ctc_capped_by_tax
        )

        return min_(maximum_refundable_ctc, amount_ctc_would_increase)

    def formula_2021(tax_unit, period, parameters):
        return aggr(tax_unit, period, ["ctc_refundable_individual_maximum"])

    formula_2022 = formula


class non_refundable_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable CTC"
    unit = USD
    documentation = (
        "The portion of the Child Tax Credit that is not refundable."
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit("ctc", period) - tax_unit("refundable_ctc", period)


c11070 = variable_alias("c11070", refundable_ctc)



from openfisca_us.model_api import *


class state_dependent_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "State dependent credit"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).states.tax.income.credits.refundable.dependent
        person = tax_unit.members
        state = person.household("state_code_str", period)
        dependent = person("is_tax_unit_dependent", period)
        age = person("age", period)
        child = age < p.child_age_limit[state]
        elderly = age >= p.elderly_age_limit[state]
        disabled = person("is_disabled", period)
        eligible = dependent & (child | elderly | disabled)
        count_eligible = tax_unit.sum(eligible)
        tax_unit_state = tax_unit.household("state_code_str", period)
        capped_eligible = min_(count_eligible, p.cap[tax_unit_state])
        return capped_eligible * p.amount[tax_unit_state]



from openfisca_us.model_api import *


class state_income_tax_rental_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax rental deduction"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        rent = tax_unit.spm_unit("rent", period)
        p = parameters(period).states.tax.income.deductions.rent
        state = tax_unit.household("state_code_str", period)
        filing_status = tax_unit("filing_status", period)
        share = p.share[state]
        cap = p.cap[state][filing_status]
        uncapped = rent * share
        return min_(uncapped, cap)



from openfisca_us.model_api import *


class basic_income(Variable):
    value_type = float
    entity = Person
    label = "Basic income"
    unit = USD
    documentation = "Total basic income payments for this person."
    definition_period = YEAR

    def formula(person, period, parameters):
        bi = parameters(period).contrib.ubi_center.basic_income
        age = person("age", period)
        return bi.amount_by_age.calc(age)



from openfisca_us.model_api import *


class state_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        before_refundable_credits = tax_unit(
            "state_income_tax_before_refundable_credits", period
        )
        refundable_credits = tax_unit(
            "state_income_tax_refundable_credits", period
        )
        return before_refundable_credits - refundable_credits



from openfisca_us.model_api import *


class dsi(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Dependent elsewhere"
    documentation = "Claimed as dependent in another tax unit"



from openfisca_us.model_api import *


class is_state_income_tax_exempt(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Exempt from state income tax"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        limit = tax_unit("state_income_tax_exempt_limit", period)
        agi = tax_unit("adjusted_gross_income", period)
        return agi <= limit



from openfisca_us.model_api import *


class ctc_maximum(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maximum CTC"
    unit = USD
    documentation = "Maximum value of the Child Tax Credit, before phaseout."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return aggr(tax_unit, period, ["ctc_individual_maximum"])



from openfisca_us.model_api import *


class pre_qbid_taxinc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Taxable income (pre-QBID)"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        uc_amount = tax_unit("tax_unit_unemployment_compensation", period)
        agi = tax_unit("adjusted_gross_income", period)
        uc_excluded = uc_amount - tax_unit(
            "tax_unit_taxable_unemployment_compensation", period
        )
        maximum_deduction = max_(
            tax_unit("c04470", period), tax_unit("standard", period)
        )
        personal_exemptions = tax_unit("c04600", period)
        return max_(
            0, agi - maximum_deduction - personal_exemptions - uc_excluded
        )


class c04800(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Taxable income"
    documentation = "Regular taxable income"
    unit = USD

    def formula(tax_unit, period, parameters):
        return max_(
            0, tax_unit("pre_qbid_taxinc", period) - tax_unit("qbided", period)
        )


taxable_income = variable_alias("taxable_income", c04800)



from openfisca_us.model_api import *


class state_limited_income_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "State limited income tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/doc/2021-schedule-nts-l-nrpy-no-tax-status-and-limited-income-credit/download"

    def formula(tax_unit, period, parameters):
        # Line numbers refer to MA's 2021 Schedule NTS-L-NR/PY.
        # Line 10.
        agi = tax_unit("adjusted_gross_income", period)
        exempt = tax_unit("is_state_income_tax_exempt", period)
        # Lines 10-12: Compute eligibility based on income limit.
        state = tax_unit.household("state_code_str", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).states.tax.income.credits.non_refundable.limited_income
        filer_limit = p.income_limit.filer[state][filing_status]
        dep_limit = p.income_limit.dependent[state][filing_status]
        dependents = tax_unit("tax_unit_dependents", period)
        limit = filer_limit + (dep_limit * dependents)
        eligible = agi <= limit
        # Line 13 ("No Tax Status" in MA).
        exempt_limit = tax_unit("state_income_tax_exempt_limit", period)
        # Line 14.
        income_for_credit = agi - exempt_limit
        # Line 15.
        tax_bc = tax_unit("state_income_tax_before_credits", period)
        # Line 16.
        tax_for_credit = p.percent[state] * income_for_credit
        # Line 17.
        return eligible * ~exempt * max_(0, tax_bc - tax_for_credit)



from openfisca_us.model_api import *


class pre_c04600(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Personal exemption before phase-out"
    reference = "https://www.law.cornell.edu/uscode/text/26/151"
    unit = USD

    def formula(tax_unit, period, parameters):
        exemption = parameters(period).irs.income.exemption
        return where(
            tax_unit("dsi", period),
            0,
            tax_unit("xtot", period) * exemption.amount,
        )


class exemption_phaseout_start(Variable):
    value_type = float
    entity = TaxUnit
    label = "Exemption phaseout start"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        return parameters(period).irs.income.exemption.phaseout.start[
            tax_unit("filing_status", period)
        ]


class c04600(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Personal exemptions after phase-out"
    unit = USD

    def formula(tax_unit, period, parameters):
        phaseout = parameters(period).irs.income.exemption.phaseout
        phaseout_start = tax_unit("exemption_phaseout_start", period)
        line_5 = max_(
            0, tax_unit("adjusted_gross_income", period) - phaseout_start
        )
        line_6 = line_5 / (2500 / tax_unit("sep", period))
        line_7 = phaseout.rate * line_6
        return tax_unit("pre_c04600", period) * (1 - line_7)



from openfisca_us.model_api import *


class c21040(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Phased-out itemized deductions"
    unit = USD
    documentation = "Itemized deductions that are phased out"

    def formula(tax_unit, period, parameters):
        nonlimited = add(tax_unit, period, ["c17000", "c20500"])
        phaseout = parameters(period).irs.deductions.itemized.phaseout
        filing_status = tax_unit("filing_status", period)
        c21060 = tax_unit("c21060", period)
        phaseout_amount_cap = phaseout.cap * max_(0, c21060 - nonlimited)
        uncapped_phaseout = (
            max_(0, tax_unit("posagi", period) - phaseout.start[filing_status])
            * phaseout.rate
        )

        return min_(
            uncapped_phaseout,
            phaseout_amount_cap,
        )


class c04470(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Itemized deductions after phase-out"
    unit = USD
    documentation = (
        "Itemized deductions after phase-out (zero for non-itemizers)"
    )

    def formula(tax_unit, period, parameters):
        return max_(0, tax_unit("c21060", period) - tax_unit("c21040", period))



from openfisca_us.model_api import *


class midr(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Separate filer itemizes"
    documentation = (
        "True if separately-filing spouse itemizes, otherwise false"
    )



from openfisca_us.model_api import *


class meets_wic_categorical_eligibility(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    documentation = (
        "Meets the program participation eligibility criteria for WIC"
    )
    label = "Meets WIC categorical (program participation) eligibility"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#d_2_A"

    def formula(spm_unit, period, parameters):
        programs = parameters(period).usda.wic.categorical_eligibility
        return np.any(
            [spm_unit(program, period) for program in programs], axis=0
        )



from openfisca_us.model_api import *


class dwks6(Variable):
    value_type = float
    entity = TaxUnit
    label = "DWKS6"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        dwks2 = tax_unit("filer_e00650", period)
        dwks3 = tax_unit("filer_e58990", period)
        # dwks4 always assumed to be zero
        dwks5 = max_(0, dwks3)
        return max_(0, dwks2 - dwks5)


class dwks9(Variable):
    value_type = float
    entity = TaxUnit
    label = "DWKS9"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p23250 = tax_unit("filer_p23250", period)
        c23650 = tax_unit("c23650", period)
        dwks7 = min_(p23250, c23650)  # SchD lines 15 and 16, respectively
        # dwks8 = min(dwks3, dwks4)
        # dwks9 = max(0., dwks7 - dwks8)
        # BELOW TWO STATEMENTS ARE UNCLEAR IN LIGHT OF dwks9=... COMMENT
        e01100 = tax_unit("filer_e01100", period)
        c24510 = where(e01100 > 0, e01100, max_(0, dwks7) + e01100)
        return max_(0, c24510 - min_(0, tax_unit("filer_e58990", period)))


class dwks10(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "DWKS10"
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )
    unit = USD

    def formula(tax_unit, period, parameters):
        dwks10_if_gains = add(tax_unit, period, ["dwks6", "dwks9"])
        dwks10_if_no_gains = max_(
            0,
            min_(
                tax_unit("filer_p23250", period),
                tax_unit("c23650", period),
            ),
        ) + tax_unit("filer_e01100", period)
        return where(
            tax_unit("hasqdivltcg", period),
            dwks10_if_gains,
            dwks10_if_no_gains,
        )


class dwks13(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "DWKS13"
    unit = USD
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )

    def formula(tax_unit, period, parameters):
        dwks1 = tax_unit("c04800", period)
        e24515 = tax_unit("filer_e24515", period)
        dwks11 = e24515 + tax_unit(
            "filer_e24518", period
        )  # Sch D lines 18 and 19, respectively
        dwks9 = tax_unit("dwks9", period)
        dwks12 = min_(dwks9, dwks11)
        dwks10 = tax_unit("dwks10", period)
        return (dwks10 - dwks12) * tax_unit("hasqdivltcg", period)


class dwks14(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "DWKS14"
    unit = USD
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )

    def formula(tax_unit, period, parameters):
        dwks1 = tax_unit("c04800", period)
        dwks13 = tax_unit("dwks13", period)
        return max_(0, dwks1 - dwks13) * tax_unit("hasqdivltcg", period)


class dwks19(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "DWKS14"
    unit = USD
    documentation = (
        "search taxcalc/calcfunctions.py for how calculated and used"
    )

    def formula(tax_unit, period, parameters):
        dwks14 = tax_unit("dwks14", period)
        capital_gains = parameters(period).irs.capital_gains.brackets
        filing_status = tax_unit("filing_status", period)
        dwks1 = tax_unit("c04800", period)
        dwks16 = min_(capital_gains.thresholds["1"][filing_status], dwks1)
        dwks17 = min_(dwks14, dwks16)
        dwks10 = tax_unit("dwks10", period)
        dwks18 = max_(0, dwks1 - dwks10)
        return max_(dwks17, dwks18) * tax_unit("hasqdivltcg", period)



from openfisca_us.model_api import *


class c07180(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Child/dependent care credit"
    unit = USD
    documentation = "Nonrefundable credit for child and dependent care expenses from Form 2441"
    reference = "https://www.law.cornell.edu/uscode/text/26/21"

    def formula(tax_unit, period, parameters):
        cdcc = parameters(period).irs.credits.cdcc
        if cdcc.refundable or cdcc.abolition:
            return 0
        else:
            c05800 = tax_unit("c05800", period)
            e07300 = tax_unit("e07300", period)
            c33200 = tax_unit("c33200", period)
            c05800_minus_e07300_capped = max_(c05800 - e07300, 0)
            return min_(c05800_minus_e07300_capped, c33200)


cdcc = variable_alias("cdcc", c07180)



from openfisca_us.model_api import *


class state_income_tax_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax exemptions"
    unit = USD
    definition_period = YEAR

    formula = sum_of_variables(
        [
            "state_income_tax_personal_exemption",
            "state_income_tax_dependent_exemption",
            "state_income_tax_aged_exemption",
            "state_income_tax_blind_exemption",
        ]
    )



from openfisca_us.model_api import *


class basic_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Basic standard deduction"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        std = parameters(period).irs.deductions.standard
        filing_status = tax_unit("filing_status", period)
        midr = tax_unit("midr", period)

        c15100_if_dsi = max_(
            std.dependent.additional_earned_income
            + tax_unit("filer_earned", period),
            std.dependent.amount,
        )
        basic_if_dsi = min_(std.amount[filing_status], c15100_if_dsi)
        basic_if_not_dsi = where(midr, 0, std.amount[filing_status])
        return where(tax_unit("dsi", period), basic_if_dsi, basic_if_not_dsi)


class aged_blind_extra_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Aged and blind standard deduction"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        std = parameters(period).irs.deductions.standard
        filing_status = tax_unit("filing_status", period)
        filing_status_type = filing_status.possible_values
        blind_head = tax_unit("blind_head", period) * 1
        blind_spouse = tax_unit("blind_spouse", period) * 1
        aged_head = (
            tax_unit("age_head", period) >= std.aged_or_blind.age_threshold
        ) * 1
        aged_spouse = (
            (filing_status == filing_status_type.JOINT)
            & (
                tax_unit("age_spouse", period)
                >= std.aged_or_blind.age_threshold
            )
        ) * 1
        num_extra_stded = blind_head + blind_spouse + aged_head + aged_spouse
        return num_extra_stded * std.aged_or_blind.amount[filing_status]


class standard(Variable):
    value_type = float
    entity = TaxUnit
    label = "Standard deduction (zero for itemizers)"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        # Calculate basic standard deduction
        basic_stded = tax_unit("basic_standard_deduction", period)
        charity = parameters(period).irs.deductions.itemized.charity
        filing_status = tax_unit("filing_status", period)
        midr = tax_unit("midr", period)
        filing_status_type = filing_status.possible_values

        # Calculate extra standard deduction for aged and blind
        extra_stded = tax_unit("aged_blind_extra_standard_deduction", period)

        # Calculate the total standard deduction
        standard = basic_stded + extra_stded
        standard = where(
            (filing_status == filing_status_type.SEPARATE) & midr, 0, standard
        )
        return standard + charity.allow_nonitemizers * min_(
            tax_unit("c19700", period), charity.nonitemizers_max
        )



from openfisca_us.model_api import *


class medicaid_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "Medicaid income"
    documentation = (
        "Modified adjusted gross income for calculating Medicaid eligibility."
    )
    unit = USD
    reference = (
        "https://www.law.cornell.edu/uscode/text/42/1396a#e_14_G",  # Medicaid law pointing to IRC
        "https://www.law.cornell.edu/uscode/text/26/36B#d_2",  # IRC defining income
    )

    def formula(spm_unit, period, parameters):
        # Medicaid law directs to use MAGI for individuals and 'household income' for families > 1 person.
        # 'Household income' is itself defined as the sum of MAGI for all individuals required to file taxes.
        person = spm_unit.members
        is_tax_unit_head = person("is_tax_unit_head", period)
        filer_medicaid_magi = is_tax_unit_head * person.tax_unit(
            "tax_unit_medicaid_magi", period
        )
        return spm_unit.sum(filer_medicaid_magi)



from openfisca_us.model_api import *


class is_on_tribal_land(Variable):
    value_type = bool
    entity = Household
    definition_period = ETERNITY
    label = "Is on tribal land"
    documentation = "Whether the household is on tribal land"



from openfisca_us.model_api import *


class is_ssi_aged_blind_disabled(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Indicates whether a person is aged, blind, or disabled for the Supplemental Security Income program"
    label = "SSI aged, blind, or disabled"

    def formula(person, period, parameters):
        aged_threshold = parameters(period).ssa.ssi.eligibility.aged_threshold
        aged = person("age", period) >= aged_threshold
        return any_(person, period, ["is_blind", "is_ssi_disabled"]) | aged



from openfisca_us.model_api import *


class phone_cost(Variable):
    value_type = float
    entity = SPMUnit
    label = "Phone cost"
    documentation = "Phone line cost for this SPM unit"
    unit = USD
    definition_period = YEAR
